/****************************************************************************\
*                                                                            *
*  fire.cpp                                                                  *
*                                                                            *
*  This demo shows explosions and fire. Written by David Freeman. 7/00       *
*  email: freeman@eng.fiu.edu                                                *
*                                                                            *
\****************************************************************************/

#include <fgwin.h>

#define KB_SPACE  57
#define KB_ESC 1

#define VBWIDTH 400
#define VBHEIGHT 300

#define NUMB_PARTICLES 3000
#define ENERGY 5

#define IRAND(min,max) ((rand() % ((max)-(min)+1)) + (min))

class PARTICLE
{
   public:
   double x;
   double y;
   double dirx;
   double diry;
   int color;
};

PARTICLE *particles[NUMB_PARTICLES];
double gravity =  0.1;
char RGBvalues[255*3];

void App_Main();
void Blit();
void Blur_Screen();
void Create_Palette();
void Create_Particles();
void Destroy_Particles();
void Draw_Particle(PARTICLE *);
void Draw_Particles();
void Explosion();
void Init_All();
void Move_Particles();
void Random_Explosion();

HWND global_window_handle = NULL;

LRESULT CALLBACK WindowProc(HWND,UINT,WPARAM,LPARAM);

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdParam, int iCmdShow)
{
   static char szAppName[] = "FGParts1";
   HWND        hWnd;
   MSG         msg;
   WNDCLASSEX  wndclass;

   wndclass.cbSize        = sizeof(wndclass);
   wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
   wndclass.lpfnWndProc   = WindowProc;
   wndclass.cbClsExtra    = 0;
   wndclass.cbWndExtra    = 0;
   wndclass.hInstance     = hInstance;
   wndclass.hIcon         = LoadIcon(NULL,IDI_APPLICATION);
   wndclass.hCursor       = LoadCursor(NULL,IDC_ARROW);
   wndclass.hbrBackground = NULL;
   wndclass.lpszMenuName  = NULL;
   wndclass.lpszClassName = szAppName;
   wndclass.hIconSm       = LoadIcon(NULL,IDI_APPLICATION);
   RegisterClassEx(&wndclass);

   hWnd = CreateWindow(szAppName, // window class name
      "Fastgraph Particle Explosion Example", // window caption
      WS_OVERLAPPEDWINDOW,     // window style
      CW_USEDEFAULT,           // initial x position
      CW_USEDEFAULT,           // initial y position
      VBWIDTH+4,               // initial x size
      VBHEIGHT+21,             // initial y size
      NULL,                    // parent window handle
      NULL,                    // window menu handle
      hInstance,               // program instance handle
      NULL);                   // creation parameters

   global_window_handle = hWnd;

   ShowWindow(hWnd,iCmdShow);
   UpdateWindow(hWnd);

   Init_All();

   while (TRUE)
   {

      if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
      {
           if (msg.message == WM_QUIT)
         break;

         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }
      App_Main();
   }
   return msg.wParam;
}

/****************************************************************************\
*                                                                            *
*  WindowProc()                                                              *
*                                                                            *
\****************************************************************************/

HDC      hDC;
HPALETTE hPal;
int      hVB;
UINT     cxClient, cyClient;

LRESULT CALLBACK WindowProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
   PAINTSTRUCT ps;

   switch (iMsg)
   {
      case WM_CREATE:
         hDC = GetDC(hWnd);
         fg_setdc(hDC);
         hPal = fg_defpal();
         fg_realize(hPal);

         fg_vbinit();
         hVB = fg_vballoc(VBWIDTH,VBHEIGHT);
         fg_vbopen(hVB);
         fg_vbcolors();
         fg_setcolor(0);
         fg_fillpage();
         Create_Palette();
         MessageBox(hWnd,"Press the space bar to launch explosives",
           "Fire Demo by David Freeman",MB_OK);
         return 0;

      case WM_PAINT:
         BeginPaint(hWnd,&ps);
         fg_vbscale(0,VBWIDTH-1,0,VBHEIGHT-1,0,cxClient-1,0,cyClient-1);
         EndPaint(hWnd,&ps);
         return 0;

      case WM_SETFOCUS:
         fg_realize(hPal);
         InvalidateRect(hWnd,NULL,TRUE);
         return 0;

      case WM_SIZE:
         cxClient = LOWORD(lParam);
         cyClient = HIWORD(lParam);
         return 0;

      case WM_DESTROY:
         // Destroy_Particles();
         fg_vbclose();
         fg_vbfree(hVB);
         fg_vbfin();

         DeleteObject(hPal);
         ReleaseDC(hWnd,hDC);
         PostQuitMessage(0);
         return 0;
   }
   return DefWindowProc(hWnd,iMsg,wParam,lParam);
}

// ****************************************************************************
void App_Main()
{
   if (fg_kbtest(KB_ESC))
   {
      PostMessage(global_window_handle, WM_DESTROY,0,0);
   }
   else if (fg_kbtest(KB_SPACE))
   {
      Random_Explosion();
   }
   Draw_Particles();
   Move_Particles();
   Blur_Screen();
   Blit();
}

// ****************************************************************************
void Blit()
{
   if ( (cxClient > VBWIDTH) || (cyClient > VBHEIGHT) )
      fg_vbscale(0,VBWIDTH-1,0,VBHEIGHT-1,0,cxClient-1,0,cyClient-1);
   else
      fg_vbpaste(0,VBWIDTH-1,0,VBHEIGHT-1,0,cyClient-1);
}

// ****************************************************************************
void Blur_Screen()
{
   int x, y, color;

   // set the left and right sides to black
   fg_setcolor(10);
   fg_rect(0,0,0,VBHEIGHT-1);
   fg_rect(VBWIDTH-1,VBWIDTH-1,0,VBHEIGHT-1);

   for (y=1; y<VBHEIGHT-2; y++)
   {
      for(x=1; x<VBWIDTH-1; x++)
      {
                       // row1            //row3               //row2
         color = (  fg_getpixel(x,y) + fg_getpixel(x,y+2) + fg_getpixel(x-1,y+1) ) / 3;
         color -= 2;

         if (color < 10)
             color = 10;

          // Set current pixel to calculated color
         fg_setcolor(color);
         fg_putpixel(x,y);
      }
   }
}

// ****************************************************************************
void Create_Particles()
{
   int i;
   PARTICLE *new_particle;

   for (i = 0; i < NUMB_PARTICLES; i++)
   {
      new_particle = new PARTICLE;
      particles[i] = new_particle;
   }
}

// ****************************************************************************
void Destroy_Particles()
{
   if (particles)
      free(particles);
}

// ****************************************************************************
void Init_All()
{
   Create_Particles();
   Explosion();
}

// ****************************************************************************
void Explosion()
{
   long i;
   double lx, ly, dist;
   double len;

   for (i = 0; i < NUMB_PARTICLES; i++)
   {
      // Set coordinates for origin of explosion
      particles[i]->x = 160.0;
      particles[i]->y = 180.0;

      // Set the x direction to either a positive or negative
      // i.e. right or left
      particles[i]->dirx = IRAND(-25,25);

      // Set the y direction to a negative 
      particles[i]->diry = IRAND(-5,0);

      // Set the energy (power) of the explosion
      dist = IRAND(1,ENERGY);

      // Store x,y directions in temp variables
      lx = particles[i]->dirx;
      ly = particles[i]->diry;

      // Use distance formula
      len = lx*lx + ly*ly;
                
      // We need to normalize the vector to give us a direction
      // vector, then find a random length of the vector, to give
      // a nice, non-uniform distribution

      if (len == 0.0)
         len = 0.0;
      else
         len = 1.0 / len;

      // Assign vector
      particles[i]->dirx *= len*dist;
      particles[i]->diry *= len*dist;

      // Set particle to brightest color
      particles[i]->color = 245;
   }
}

// ****************************************************************************
void Draw_Particle(PARTICLE *single_particle)
{
   int x, y;

   x = (int) (single_particle->x + 0.5);
   y = (int) (single_particle->y + 0.5);

   fg_setcolor(single_particle->color);
   fg_point(x,y);
}

// ****************************************************************************
void Draw_Particles()
{
   int i;
   PARTICLE *parts;

   parts = *particles;
   for(i = 0; i < NUMB_PARTICLES; i++)
   {
      Draw_Particle(parts);
      parts++;
   }
}

// ****************************************************************************
void Move_Particles()
{
   int i;
   PARTICLE *part;

   part = *particles;

   for (i = 0; i < NUMB_PARTICLES; i++)
   {
      // Move particle according to vector direction
      part->x += part->dirx;
      part->y += part->diry;

      // Check if the particle has collided with any of the screen
      // top or bottom edges. If so, reflect the direction, and
      // lose energy. If not, then just add gravity

      if (part->y >= VBHEIGHT-2)
      {
         part->y = VBHEIGHT-2;
         part->dirx /= 4;
         part->diry = -part->diry / 2;
      }
      else if (part->y < 1)
      {
         part->y = 1;
         part->dirx /= 4;
         part->diry = -part->diry / 2;
      }
      else
      {
         part->diry += gravity;
      }

      if (part->x < 0)
      {
         part->x = 1;
         part->dirx = -part->dirx / 2;
         part->diry /= 4;
      }
      else if (part->x > VBWIDTH-1)
      {
         part->x = VBWIDTH-1;
         part->dirx = -part->dirx / 2;
         part->diry /= 4;
      }

     // If the particle is near the bottom of the screen, we
     // make its color random, to simulate the effect of the
     // particle dying, things catching fire, that kind of stuff
     if (part->y >= VBHEIGHT-2)
        part->color = IRAND(10,246);
     else
        part->color = max(part->color-4,10);

     part++;
  }
}

// ****************************************************************************
void Random_Explosion()
{
   int i;
   double len, lx, ly, dist;
   int x, y;

   x = IRAND(0,VBWIDTH-1);
   y = IRAND(0,VBHEIGHT-2);

   for (i = 0; i < NUMB_PARTICLES; i++)
   {
      // Set coordinates for origin of explosion
      particles[i]->x = x;
      particles[i]->y = y;

      // Set the x and y directions to either a positive or negative
      // i.e. right/left or up/down
      particles[i]->dirx = IRAND(-1,1);
      particles[i]->diry = IRAND(-1,1);

      // Set the energy (power) of the explosion
      dist = IRAND(1,ENERGY);

      // Store x,y directions in temp variables
      lx = particles[i]->dirx;
      ly = particles[i]->diry;

      // Use distance formula
      len = lx*lx + ly*ly;
                
      if (len != 0.0)
          len = 1.0 / len;

      // We need to normalize the vector to give us a direction
      // vector, then find a random length of the vector, to give
      // a nice, non-uniform distribution

      particles[i]->dirx *= len*dist;
      particles[i]->diry *= len*dist;

      // Set particle to brightest color
      particles[i]->color = 245;
   }
}

// ****************************************************************************
void Create_Palette()
{
   int i;
   // ideally this setups a palette that is supposed to fade from red->yellow->wht
   // this is supposed to simuliate an explosion color spread.

   // palette entry 10 is black
   fg_setrgb(10, 0,0,0);

   // set up reds
   for (i = 11; i <= 63; i++)
   {
      fg_setrgb(i, i+192, i+64, 32);
   }

   // set up yellow and orange
   for (i = 64; i <= 127; i++)
   {
      fg_setrgb(i, 255, 2*i, 32);
   }

   // setup yellow to white
   for (i = 128; i <= 245; i++)
   {
      fg_setrgb(i, 255, 255, i+10);
   }
}
